from __future__ import (absolute_import, division, print_function)

from __future__ import unicode_literals
# this example reads today's numerical weather forecasts
# from the NOAA OpenDAP servers and makes a multi-panel plot.
# This version demonstrates the use of the AxesGrid toolkit.
import numpy as np
import matplotlib.pyplot as plt
import sys
import numpy.ma as ma
import datetime
from mpl_toolkits.basemap import Basemap, addcyclic
from mpl_toolkits.axes_grid1 import AxesGrid
from netCDF4 import Dataset as NetCDFFile, num2date


# today's date is default.
if len(sys.argv) > 1:
    YYYYMMDD = sys.argv[1]
else:
    YYYYMMDD = datetime.datetime.today().strftime('%Y%m%d')

# set OpenDAP server URL.
try:
    URLbase="http://nomads.ncep.noaa.gov:9090/dods/gfs/gfs"
    URL=URLbase+YYYYMMDD+'/gfs_00z'
    print(URL)
    data = NetCDFFile(URL)
except:
    msg = """
opendap server not providing the requested data.
Try another date by providing YYYYMMDD on command line."""
    raise IOError(msg)


# read lats,lons,times.

print(data.variables.keys())
latitudes = data.variables['lat']
longitudes = data.variables['lon']
fcsttimes = data.variables['time']
times = fcsttimes[0:6] # first 6 forecast times.
ntimes = len(times)
# convert times for datetime instances.
fdates = num2date(times,units=fcsttimes.units,calendar='standard')
# make a list of YYYYMMDDHH strings.
verifdates = [fdate.strftime('%Y%m%d%H') for fdate in fdates]
# convert times to forecast hours.
fcsthrs = []
for fdate in fdates:
    fdiff = fdate-fdates[0]
    fcsthrs.append(fdiff.days*24. + fdiff.seconds/3600.)
print(fcsthrs)
print(verifdates)
lats = latitudes[:]
nlats = len(lats)
lons1 = longitudes[:]
nlons = len(lons1)

# unpack 2-meter temp forecast data.

t2mvar = data.variables['tmp2m']

# create figure, set up AxesGrid.
fig=plt.figure(figsize=(6,8))
grid = AxesGrid(fig, [0.05,0.01,0.9,0.9],
                nrows_ncols=(3, 2),
		axes_pad=0.25,
		cbar_mode='single',
		cbar_pad=0.3,
		cbar_size=0.1,
                cbar_location='top',
                share_all=True,
		)

# create Basemap instance for Orthographic projection.
m = Basemap(lon_0=-90,lat_0=60,projection='ortho')
# add wrap-around point in longitude.
t2m = np.zeros((ntimes,nlats,nlons+1),np.float32)
for nt in range(ntimes):
    t2m[nt,:,:], lons = addcyclic(t2mvar[nt,:,:], lons1)
# convert to celsius.
t2m = t2m-273.15
# contour levels
clevs = np.arange(-30,30.1,2.)
lons, lats = np.meshgrid(lons, lats)
x, y = m(lons, lats)
# make subplots.
for nt,fcsthr in enumerate(fcsthrs):
    ax = grid[nt]
    m.ax = ax
    cs = m.contourf(x,y,t2m[nt,:,:],clevs,cmap=plt.cm.jet,extend='both')
    m.drawcoastlines(linewidth=0.5)
    m.drawcountries()
    m.drawparallels(np.arange(-80,81,20))
    m.drawmeridians(np.arange(0,360,20))
    # panel title
    ax.set_title('%d-h forecast valid '%fcsthr+verifdates[nt],fontsize=9)
# figure title
plt.figtext(0.5,0.95,
            "2-m temp (\N{DEGREE SIGN}C) forecasts from %s"%verifdates[0],
            horizontalalignment='center',fontsize=14)
# a single colorbar.
cbar = fig.colorbar(cs, cax=grid.cbar_axes[0], orientation='horizontal')
plt.show()



from __future__ import (absolute_import, division, print_function)

from mpl_toolkits.basemap import Basemap
import numpy as np
import matplotlib.pyplot as plt
import pickle, time

# create figure with aqua background (will be oceans)
fig = plt.figure()

# create Basemap instance. Use 'high' resolution coastlines.
t1 = time.clock()
#m = Basemap(llcrnrlon=-10.5,llcrnrlat=49.5,urcrnrlon=3.5,urcrnrlat=59.5,
#            resolution='h',projection='tmerc',lon_0=-4,lat_0=0)
m = Basemap(width=920000,height=1100000,
            resolution='f',projection='tmerc',lon_0=-4.2,lat_0=54.6)
# make sure countries and rivers are loaded
m.drawcountries()
m.drawrivers()
print(time.clock()-t1,' secs to create original Basemap instance')

# pickle the class instance.
pickle.dump(m,open('map.pickle','wb'),-1)

# clear the figure
plt.clf()
# read pickle back in and plot it again (should be much faster).
t1 = time.clock()
m2 = pickle.load(open('map.pickle','rb'))
# draw coastlines and fill continents.
m.drawcoastlines()
# fill continents and lakes
m.fillcontinents(color='coral',lake_color='aqua')
# draw political boundaries.
m.drawcountries(linewidth=1)
# fill map projection region light blue (this will
# paint ocean areas same color as lakes).
m.drawmapboundary(fill_color='aqua')
# draw major rivers.
m.drawrivers(color='b')
print(time.clock()-t1,' secs to plot using using a pickled Basemap instance')
# draw parallels
circles = np.arange(48,65,2).tolist()
m.drawparallels(circles,labels=[1,1,0,0])
# draw meridians
meridians = np.arange(-12,13,2)
m.drawmeridians(meridians,labels=[0,0,1,1])
plt.title("High-Res British Isles",y=1.04)
plt.show()



from netCDF4 import Dataset, num2date
import time, calendar, datetime, numpy
from mpl_toolkits.basemap import Basemap
import matplotlib.pyplot as plt
import urllib, os
# data downloaded from the form at
# http://coastwatch.pfeg.noaa.gov/erddap/tabledap/apdrcArgoAll.html 
filename, headers = urllib.urlretrieve('http://coastwatch.pfeg.noaa.gov/erddap/tabledap/apdrcArgoAll.nc?longitude,latitude,time&longitude>=0&longitude<=360&latitude>=-90&latitude<=90&time>=2010-01-01&time<=2010-01-08&distinct()')
dset = Dataset(filename)
lats = dset.variables['latitude'][:]
lons = dset.variables['longitude'][:]
time = dset.variables['time']
times = time[:]
t1 = times.min(); t2 = times.max()
date1 = num2date(t1, units=time.units)
date2 = num2date(t2, units=time.units)
dset.close()
os.remove(filename)
# draw map with markers for float locations
m = Basemap(projection='hammer',lon_0=180)
x, y = m(lons,lats)
m.drawmapboundary(fill_color='#99ffff')
m.fillcontinents(color='#cc9966',lake_color='#99ffff')
m.scatter(x,y,3,marker='o',color='k')
plt.title('Locations of %s ARGO floats active between %s and %s' %\
        (len(lats),date1,date2),fontsize=12)
plt.show()



from mpl_toolkits.basemap import Basemap
from netCDF4 import Dataset, date2index
import numpy as np
import matplotlib.pyplot as plt
from datetime import datetime
date = datetime(2007,12,15,0) # date to plot.
# open dataset.
dataset = \
Dataset('http://www.ncdc.noaa.gov/thredds/dodsC/OISST-V2-AVHRR_agg')
timevar = dataset.variables['time']
timeindex = date2index(date,timevar) # find time index for desired date.
# read sst.  Will automatically create a masked array using
# missing_value variable attribute. 'squeeze out' singleton dimensions.
sst = dataset.variables['sst'][timeindex,:].squeeze()
# read ice.
ice = dataset.variables['ice'][timeindex,:].squeeze()
# read lats and lons (representing centers of grid boxes).
lats = dataset.variables['lat'][:]
lons = dataset.variables['lon'][:]
lons, lats = np.meshgrid(lons,lats)
# create figure, axes instances.
fig = plt.figure()
ax = fig.add_axes([0.05,0.05,0.9,0.9])
# create Basemap instance.
# coastlines not used, so resolution set to None to skip
# continent processing (this speeds things up a bit)
m = Basemap(projection='kav7',lon_0=0,resolution=None)
# draw line around map projection limb.
# color background of map projection region.
# missing values over land will show up this color.
m.drawmapboundary(fill_color='0.3')
# plot sst, then ice with pcolor
im1 = m.pcolormesh(lons,lats,sst,shading='flat',cmap=plt.cm.jet,latlon=True)
im2 = m.pcolormesh(lons,lats,ice,shading='flat',cmap=plt.cm.gist_gray,latlon=True)
# draw parallels and meridians, but don't bother labelling them.
m.drawparallels(np.arange(-90.,99.,30.))
m.drawmeridians(np.arange(-180.,180.,60.))
# add colorbar
cb = m.colorbar(im1,"bottom", size="5%", pad="2%")
# add a title.
ax.set_title('SST and ICE analysis for %s'%date)
plt.show()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Víctor Molina García

# GeosLibrary.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# GeosLibrary.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with GeosLibrary.py. If not, see <https://www.gnu.org/licenses/>.
#
"""Helper class to download and build the GEOS library."""

import io
import os
import ssl
import glob
import shutil
import struct
import tempfile
import contextlib
import subprocess
import datetime as dt
from zipfile import ZipFile
try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen


URL_DATETIME_FMT = "%a, %d %b %Y %H:%M:%S GMT"
GEOS_BASEURL = "https://github.com/libgeos/geos/archive/refs/tags"


class GeosLibrary(object):
    """Helper class to download, build and install GEOS."""

    def __init__(self, version, root=None):
        """Initialise a new :class:`GeosLibrary` instance."""

        self.version_tuple = tuple(map(int, version.split(".")))

        if root is None:
            self.temp = True
            self.root = tempfile.mkdtemp(prefix="tmp_geoslibrary_")
        else:
            self.temp = False
            self.root = root
            try:
                os.makedirs(self.root)
            except OSError:
                pass

    def __del__(self):
        """Clean up after :class:`GeosLibrary` destruction."""

        if self.temp:
            try:
                shutil.rmtree(self.root)
            except OSError:
                pass

    @property
    def version(self):
        """GEOS library version in string format."""

        return ".".join(map(str, self.version_tuple))

    def download(self):
        """Download GEOS zip source code into :class:`GeosLibrary` root."""

        # Define download link.
        link = "{0}/{1}.zip".format(GEOS_BASEURL, self.version)
        suffix = os.path.splitext(link)[-1]

        # Define output path.
        zipname = "geos-{0}".format(link.rsplit("/", 1)[-1])
        zippath = os.path.join(self.root, zipname)

        # Handle creation of the HTTP request.
        kwargs = {}
        if hasattr(ssl, "SSLContext") and hasattr(ssl, "PROTOCOL_TLSv1_2"):
            kwargs.update(context=ssl.SSLContext(ssl.PROTOCOL_TLSv1_2))
        try:
            conn = urlopen(link, **kwargs)
        except TypeError:
            # Fallback if `urlopen` does not accept context.
            conn = urlopen(link)

        with contextlib.closing(conn):
            # Try to get the file timestamp from the HTTP request header.
            date = conn.headers.get("Last-Modified")
            if date is not None:
                date = dt.datetime.strptime(date, URL_DATETIME_FMT)
                date = date.replace(tzinfo=dt.timezone.utc)
                date = date.timestamp()
            with tempfile.NamedTemporaryFile(suffix=suffix) as tmpzipobj:
                # Copy the buffer into a temporary file.
                shutil.copyfileobj(conn, tmpzipobj)
                # Move the file descriptor pointer to the beginning and
                # simply copy it to the final destination.
                tmpzipobj.seek(0)
                with open(zippath, "wb") as zipobj:
                    shutil.copyfileobj(tmpzipobj, zipobj)
            # Assign the timestamps to the final file if available.
            if date is not None:
                os.utime(zippath, (date, date))

    def extract(self, overwrite=True):
        """Decompress GEOS zip source code into :class:`GeosLibrary` root."""

        # Download zip file if not present.
        zippath = os.path.join(self.root, "geos-{0}.zip".format(self.version))
        if not os.path.exists(zippath):
            self.download()

        # Remove destination folder if present and requested.
        zipfold = os.path.join(self.root, "geos-{0}".format(self.version))
        if os.path.exists(zipfold):
            if not overwrite:
                raise OSError("folder '{0}' already exists".format(zipfold))
            else:
                shutil.rmtree(zipfold)

        # Decompress zip file.
        with contextlib.closing(ZipFile(zippath, "r")) as fd:
            fd.extractall(self.root)

        # Ensure that GEOS internal sh scripts can be executed.
        for path in sorted(glob.glob(os.path.join(zipfold, "tools", "*.sh"))):
            os.chmod(path, 0o755)

        # Patch CMakeLists so that libgeos_c.so does not depend on libgeos.so.
        cmakefile = os.path.join(zipfold, "capi", "CMakeLists.txt")
        with io.open(cmakefile, "r", encoding="utf-8") as fd:
            lines = fd.readlines()
        with io.open(cmakefile, "wb") as fd:
            oldtext = "target_link_libraries(geos_c geos)"
            newtext = "target_link_libraries(geos_c geos-static)"
            for line in lines:
                fd.write(line.replace(oldtext, newtext).encode())

        # The SVN revision file is not created on the fly before 3.6.0.
        svn_hfile = os.path.join(zipfold, "geos_svn_revision.h")
        if self.version_tuple < (3, 6, 0) and not os.path.exists(svn_hfile):
            with io.open(svn_hfile, "wb") as fd:
                text = "#define GEOS_SVN_REVISION 0"
                fd.write(text.encode())

    def build(self, installdir=None, njobs=1):
        """Build and install GEOS from source."""

        # Download and extract zip file if not present.
        zipfold = os.path.join(self.root, "geos-{0}".format(self.version))
        self.extract(overwrite=True)

        # Define build directory.
        builddir = os.path.join(zipfold, "build")

        # Define installation directory.
        if installdir is None:
            installdir = os.path.expanduser("~/.local/share/libgeos")
        installdir = os.path.abspath(installdir)

        # Define configure options.
        config_opts = [
            "-DCMAKE_INSTALL_PREFIX={0}".format(installdir),
            "-DGEOS_ENABLE_TESTS=OFF",
            "-DCMAKE_BUILD_TYPE=Release",
        ]
        if os.name == "nt" and self.version_tuple < (3, 6, 0):
            config_opts = ["-G", "NMake Makefiles"] + config_opts

        # Define build options.
        build_env = os.environ.copy()
        build_opts = [
            "--config", "Release",
            "--target", "install",
        ]
        if os.name != "nt":
            build_env["MAKEFLAGS"] = "-j {0:d}".format(njobs)
        elif self.version_tuple < (3, 6, 0):
            win64 = (8 * struct.calcsize("P") == 64)
            build_opts.extend([
                "--",
                "WIN64={0}".format("YES" if win64 else "NO"),
                "BUILD_BATCH={0}".format("YES" if njobs > 1 else "NO")
            ])
        else:
            build_opts = ["-j", "{0:d}".format(njobs)] + build_opts

        # Now move to the GEOS source code folder and build with CMake.
        cwd = os.getcwd()
        try:
            # Ensure that the build directory exists.
            try:
                os.makedirs(builddir)
            except OSError:
                pass
            os.chdir(builddir)
            # Call cmake configure.
            subprocess.call(["cmake", ".."] + config_opts)
            # Ensure that the install directory exists.
            try:
                os.makedirs(installdir)
            except OSError:
                pass
            # Call cmake build and install.
            subprocess.call(["cmake", "--build", "."] + build_opts,
                            env=build_env)
        finally:
            os.chdir(cwd)



import sys
import numpy

__version__ = "0.3"

# need some python C API functions for strings.
cdef extern from "Python.h":
    object PyBytes_FromString(char *)

# taken from numpy.pxi in numpy 1.0rc2.
cdef extern from "numpy/arrayobject.h":
    ctypedef int npy_intp 
    ctypedef extern class numpy.ndarray [object PyArrayObject]:
        cdef char *data
        cdef int nd
        cdef npy_intp *dimensions
        cdef npy_intp *strides
        cdef object base
#       cdef dtype descr
        cdef int flags
    npy_intp PyArray_SIZE(ndarray arr)
    npy_intp PyArray_ISCONTIGUOUS(ndarray arr)
    void import_array()

# Initialize numpy
import_array()

# GENERAL NOTES:
#
#	- Remember to call initGEOS() before any use of this library's
#	  functions, and call finishGEOS() when done.
#
#	- Currently you have to explicitly GEOSGeom_destroy() all
#	  GEOSGeom objects to avoid memory leaks, and to free()
#	  all returned char * (unless const). This might change
#	  before first release to ensure greater API stability.

cdef extern from "geos_c.h":
# Supported geometry type IDs
    cdef enum:
        GEOS_POINT
        GEOS_LINESTRING
        GEOS_LINEARRING
        GEOS_POLYGON
        GEOS_MULTIPOINT
        GEOS_MULTILINESTRING
        GEOS_MULTIPOLYGON
        GEOS_GEOMETRYCOLLECTION
        GEOS_VERSION_MAJOR
    ctypedef struct GEOSGeom:
        pass
    ctypedef struct GEOSCoordSeq:
        pass
    ctypedef void (*GEOSMessageHandler)(char *fmt, char *list)
    char *GEOSversion()
    void initGEOS(GEOSMessageHandler notice_function, GEOSMessageHandler error_function)
    void finishGEOS()
    GEOSCoordSeq *GEOSCoordSeq_create(unsigned int size, unsigned int dims)
    void GEOSCoordSeq_destroy(GEOSCoordSeq* s)
    int GEOSCoordSeq_setX(GEOSCoordSeq* s,unsigned int idx, double val)
    int GEOSCoordSeq_setY(GEOSCoordSeq* s,unsigned int idx, double val)
    int GEOSCoordSeq_getX(GEOSCoordSeq* s, unsigned int idx, double *val)
    int GEOSCoordSeq_getY(GEOSCoordSeq* s, unsigned int idx, double *val)
    GEOSGeom *GEOSUnion(GEOSGeom* g1, GEOSGeom* g2)
    GEOSGeom *GEOSUnaryUnion(GEOSGeom* g1)
    GEOSGeom *GEOSEnvelope(GEOSGeom* g1)
    GEOSGeom *GEOSConvexHull(GEOSGeom* g1)
    GEOSGeom *GEOSGeom_createPoint(GEOSCoordSeq* s)
    GEOSGeom *GEOSGeom_createLineString(GEOSCoordSeq* s)
    GEOSGeom *GEOSGeom_createPolygon(GEOSGeom* shell, GEOSGeom** holes, unsigned int nholes)
    GEOSGeom  *GEOSGeom_createLinearRing(GEOSCoordSeq* s)
    void GEOSGeom_destroy(GEOSGeom* g)
# Topology operations - return NULL on exception.
    GEOSGeom *GEOSIntersection(GEOSGeom* g1, GEOSGeom* g2)
    GEOSGeom *GEOSSimplify(GEOSGeom* g1, double tolerance)
    GEOSGeom *GEOSBuffer(GEOSGeom* g1, double width, int quadsegs)
    GEOSGeom *GEOSTopologyPreserveSimplify(GEOSGeom* g1, double tolerance)
# Binary/Unary predicate - return 2 on exception, 1 on true, 0 on false
    char GEOSIntersects(GEOSGeom* g1, GEOSGeom* g2)
    char GEOSWithin(GEOSGeom* g1, GEOSGeom* g2)
    char GEOSContains(GEOSGeom* g1, GEOSGeom* g2)
    char GEOSisEmpty(GEOSGeom* g1)
    char GEOSisValid(GEOSGeom* g1)
    char GEOSisSimple(GEOSGeom* g1)
    char GEOSisRing(GEOSGeom* g1)
#  Geometry info
    char *GEOSGeomType(GEOSGeom* g1)
    int GEOSGeomTypeId(GEOSGeom* g1)
# Functions: Return 0 on exception, 1 otherwise 
    int GEOSArea(GEOSGeom* g1, double *area)
    int GEOSLength(GEOSGeom* g1, double *length)
# returns -1 on error and 1 for non-multi geoms
    int GEOSGetNumGeometries(GEOSGeom* g1)
# Return NULL on exception, Geometry must be a Collection.
# Returned object is a pointer to internal storage:
# it must NOT be destroyed directly.
    GEOSGeom  *GEOSGetGeometryN(GEOSGeom* g, int n)
    int GEOSGetNumInteriorRings(GEOSGeom* g1)
# Return NULL on exception, Geometry must be a Polygon.
# Returned object is a pointer to internal storage:
# it must NOT be destroyed directly.
    GEOSGeom  *GEOSGetExteriorRing(GEOSGeom* g)
# Return NULL on exception.
# Geometry must be a LineString, LinearRing or Point.
    GEOSCoordSeq  *GEOSGeom_getCoordSeq(GEOSGeom* g)
    int GEOSCoordSeq_getSize(GEOSCoordSeq *s, unsigned int *size)

cdef void notice_h(char *fmt, char*msg):
    pass
    #format = PyBytes_FromString(fmt)
    #message = PyBytes_FromString(msg)
    #try:
    #    warn_msg = format % message
    #except:
    #    warn_msg = format
    #sys.stdout.write('GEOS_NOTICE: %s\n' % warn_msg)

cdef void error_h(char *fmt, char*msg):
    format = PyBytes_FromString(fmt)
    message = PyBytes_FromString(msg)
    try:
        warn_msg = format % message
    except:
        warn_msg = format
    sys.stderr.write('GEOS_ERROR: %s\n' % warn_msg)

# check library version
cdef geos_version():
    return PyBytes_FromString(GEOSversion())
__geos_version__ = geos_version() # module variable.
__geos_major_version__ = GEOS_VERSION_MAJOR
#if __geos_version__ != "2.2.3-CAPI-1.1.1":
#     raise ValueError('version 2.2.3 of the geos library is required')

# intialize GEOS (parameters are notice and error function callbacks).
initGEOS(notice_h, error_h)

cdef class BaseGeometry:
    cdef GEOSGeom *_geom
    cdef unsigned int _npts
    cdef public object boundary

    def is_valid(self):
        cdef char valid
        valid = GEOSisValid(self._geom)
        if valid:
            return True
        else:
            return False

    def geom_type(self):
        return PyBytes_FromString(GEOSGeomType(self._geom))

    def within(self, BaseGeometry geom):
        cdef GEOSGeom *g1
        cdef GEOSGeom *g2
        cdef char answer
        g1 = self._geom
        g2 = geom._geom
        answer =  GEOSWithin(g1, g2)
        if answer:
            return True
        else:
            return False

    def union(self, BaseGeometry geom):
        cdef GEOSGeom *g1
        cdef GEOSGeom *g2
        cdef GEOSGeom *g3
        cdef GEOSGeom *gout
        cdef int numgeoms, i, typeid
        g1 = self._geom
        g2 = geom._geom
        g3 = GEOSUnion(g1, g2)
        typeid = GEOSGeomTypeId(g3)
        if typeid == GEOS_POLYGON:
            b = _get_coords(g3)
            p = Polygon(b)
        elif typeid == GEOS_LINESTRING:
            b = _get_coords(g3)
            p = LineString(b)
        # for multi-geom structures, just return first one.
        elif typeid == GEOS_MULTIPOLYGON:
            numgeoms = GEOSGetNumGeometries(g3)
            gout = GEOSGetGeometryN(g3, 0)
            b = _get_coords(gout)
            p = Polygon(b)
        elif typeid == GEOS_MULTILINESTRING:
            numgeoms = GEOSGetNumGeometries(g3)
            gout = GEOSGetGeometryN(g3, 0)
            b = _get_coords(gout)
            p = LineString(b)
        else:
            type = PyBytes_FromString(GEOSGeomType(g3))
            raise NotImplementedError("unions of type '%s' not yet implemented" % (type))
        GEOSGeom_destroy(g3)
        return p

    def simplify(self, tol):
        cdef GEOSGeom *g1
        cdef GEOSGeom *g3
        cdef GEOSGeom *gout
        cdef double tolerance
        cdef int numgeoms, i, typeid
        g1 = self._geom
        tolerance = tol
        g3 = GEOSSimplify(g1,tolerance)
        typeid = GEOSGeomTypeId(g3)
        if typeid == GEOS_POLYGON:
            b = _get_coords(g3)
            p = Polygon(b)
        elif typeid == GEOS_LINESTRING:
            b = _get_coords(g3)
            p = LineString(b)
        # for multi-geom structures, just return first one.
        elif typeid == GEOS_MULTIPOLYGON:
            numgeoms = GEOSGetNumGeometries(g3)
            gout = GEOSGetGeometryN(g3, 0)
            b = _get_coords(gout)
            p = Polygon(b)
        elif typeid == GEOS_MULTILINESTRING:
            numgeoms = GEOSGetNumGeometries(g3)
            gout = GEOSGetGeometryN(g3, 0)
            b = _get_coords(gout)
            p = LineString(b)
        else:
            type = PyBytes_FromString(GEOSGeomType(g3))
            raise NotImplementedError("intersections of type '%s' not yet implemented" % (type))
        GEOSGeom_destroy(g3)
        return p

    def fix(self):
        cdef GEOSGeom *g1
        cdef GEOSGeom *g3
        cdef GEOSGeom *gout
        cdef int numgeoms, i, typeid
        g1 = self._geom
        g3 = GEOSBuffer(g1, 0., 0)
        typeid = GEOSGeomTypeId(g3)
        if typeid == GEOS_POLYGON:
            b = _get_coords(g3)
            p = Polygon(b)
        elif typeid == GEOS_LINESTRING:
            b = _get_coords(g3)
            p = LineString(b)
        # for multi-geom structures, just return first one.
        elif typeid == GEOS_MULTIPOLYGON:
            numgeoms = GEOSGetNumGeometries(g3)
            gout = GEOSGetGeometryN(g3, 0)
            b = _get_coords(gout)
            p = Polygon(b)
        elif typeid == GEOS_MULTILINESTRING:
            numgeoms = GEOSGetNumGeometries(g3)
            gout = GEOSGetGeometryN(g3, 0)
            b = _get_coords(gout)
            p = LineString(b)
        else:
            type = PyBytes_FromString(GEOSGeomType(g3))
            raise NotImplementedError("intersections of type '%s' not yet implemented" % (type))
        GEOSGeom_destroy(g3)
        return p

    def intersects(self, BaseGeometry geom):
        cdef GEOSGeom *g1
        cdef GEOSGeom *g2
        cdef char answer
        g1 = self._geom
        g2 = geom._geom
        answer =  GEOSIntersects(g1, g2)
        if answer:
            return True
        else:
            return False

    def intersection(self, BaseGeometry geom):
        cdef GEOSGeom *g1
        cdef GEOSGeom *g2
        cdef GEOSGeom *g3
        cdef GEOSGeom *gout
        cdef char answer
        cdef int numgeoms, i, typeid
        g1 = self._geom
        g2 = geom._geom
        g3 =  GEOSIntersection(g1, g2)
        typeid = GEOSGeomTypeId(g3)
        if typeid == GEOS_POLYGON:
            b = _get_coords(g3)
            p = Polygon(b)
            pout = [p]
        elif typeid == GEOS_LINESTRING:
            b = _get_coords(g3)
            p = LineString(b)
            pout = [p]
        elif typeid == GEOS_MULTIPOLYGON:
            numgeoms = GEOSGetNumGeometries(g3)
            pout = []
            for i from 0 <= i < numgeoms:
                gout = GEOSGetGeometryN(g3, i)
                b = _get_coords(gout)
                p = Polygon(b)
                pout.append(p)
        elif typeid == GEOS_MULTILINESTRING:
            numgeoms = GEOSGetNumGeometries(g3)
            pout = []
            for i from 0 <= i < numgeoms:
                gout = GEOSGetGeometryN(g3, i)
                b = _get_coords(gout)
                p = LineString(b)
                pout.append(p)
        else:
            #type = PyBytes_FromString(GEOSGeomType(g3))
            #raise NotImplementedError("intersections of type '%s' not yet implemented" % (type))
            GEOSGeom_destroy(g3)
            return []
        GEOSGeom_destroy(g3)
        return pout

    def get_coords(self):
        return _get_coords(self._geom)

    def __dealloc__(self):
        """destroy GEOS geometry"""
        GEOSGeom_destroy(self._geom)

    def __reduce__(self):
        """special method that allows geos instance to be pickled"""
        return (self.__class__,(self.boundary,))
    
cdef class Polygon(BaseGeometry):

    def __init__(self, ndarray b):
        cdef unsigned int M, m, i
        cdef double dx, dy
        cdef double *bbuffer
        cdef GEOSCoordSeq *cs
        cdef GEOSGeom *lr



        # make sure data is contiguous.
        # if not, make a local copy.
        if not PyArray_ISCONTIGUOUS(b):
            b = b.copy()

        m = b.shape[0]
       
        # Add closing coordinates to sequence?
        if m > 0 and (b[-1,0] != b[0,0] or b[-1,1] != b[0,1]):
            M = m + 1
        else:
            M = m
        self._npts = M

        # Create a coordinate sequence
        cs = GEOSCoordSeq_create(M, 2)

        # add to coordinate sequence
        bbuffer = <double *>b.data
        for i from 0 <= i < m:
            dx = bbuffer[2*i]
            dy = bbuffer[2*i+1]
            # Because of a bug in the GEOS C API, 
            # always set X before Y
            GEOSCoordSeq_setX(cs, i, dx)
            GEOSCoordSeq_setY(cs, i, dy)

        # Add closing coordinates to sequence?
        if M > m:
            dx = bbuffer[0]
            dy = bbuffer[1]
            GEOSCoordSeq_setX(cs, M-1, dx)
            GEOSCoordSeq_setY(cs, M-1, dy)

        # create LinearRing
        lr = GEOSGeom_createLinearRing(cs)

        # create Polygon from LinearRing (assuming no holes)
        self._geom = GEOSGeom_createPolygon(lr,NULL,0)
        self.boundary = b


    def area(self):
        cdef double area
        GEOSArea(self._geom, &area)
        return area

cdef class LineString(BaseGeometry):
    def __init__(self, ndarray b):
        cdef double dx, dy
        cdef GEOSCoordSeq *cs
        cdef int i, M
        cdef double *bbuffer

        # make sure data is contiguous.
        # if not, make a local copy.
        if not PyArray_ISCONTIGUOUS(b):
            b = b.copy()

        M = b.shape[0]
        self._npts = M

        # Create a coordinate sequence
        cs = GEOSCoordSeq_create(M, 2)

        # add to coordinate sequence
        bbuffer = <double *>b.data
        for i from 0 <= i < M:
            dx = bbuffer[2*i]
            dy = bbuffer[2*i+1]
            # Because of a bug in the GEOS C API, 
            # always set X before Y
            GEOSCoordSeq_setX(cs, i, dx)
            GEOSCoordSeq_setY(cs, i, dy)

        # create LineString
        self._geom = GEOSGeom_createLineString(cs)
        self.boundary = b

cdef class Point(BaseGeometry):
    cdef public x,y
    def __init__(self, b):
        cdef double dx, dy
        cdef GEOSCoordSeq *cs
        # Create a coordinate sequence
        cs = GEOSCoordSeq_create(1, 2)
        dx = b[0]; dy = b[1]
        GEOSCoordSeq_setX(cs, 0, dx)
        GEOSCoordSeq_setY(cs, 0, dy)
        self._geom = GEOSGeom_createPoint(cs)
        self._npts = 1
        self.boundary = b

cdef _get_coords(GEOSGeom *geom):
    cdef GEOSCoordSeq *cs
    cdef GEOSGeom *lr
    cdef unsigned int i, M
    cdef double dx, dy
    cdef ndarray b
    cdef double *bbuffer
    if GEOSGeomTypeId(geom) == GEOS_POLYGON:
        lr = GEOSGetExteriorRing(geom)
        cs = GEOSGeom_getCoordSeq(lr)
    else:
        cs = GEOSGeom_getCoordSeq(geom)
    GEOSCoordSeq_getSize(cs, &M)
    b = numpy.empty((M,2), numpy.float64)
    bbuffer = <double *>b.data
    for i from 0 <= i < M:
        GEOSCoordSeq_getX(cs, i, &dx)
        GEOSCoordSeq_getY(cs, i, &dy)
        bbuffer[2*i] = dx
        bbuffer[2*i+1] = dy
    return b



from __future__ import (absolute_import, division, print_function)

from matplotlib.mlab import prctile_rank
import matplotlib.pyplot as plt
from mpl_toolkits.basemap import Basemap as Basemap

# cities colored by population rank.

m = Basemap()
shp_info = m.readshapefile('cities','cities')
x, y = zip(*m.cities)
pop = []
for item in m.cities_info:
    population = item['POPULATION']
    if population < 0: continue # population missing
    pop.append(population)
popranks = prctile_rank(pop,100)
colors = []
for rank in popranks:
    colors.append(plt.cm.jet(float(rank)/100.))
m.drawcoastlines()
m.fillcontinents()
m.scatter(x,y,25,colors,marker='o',edgecolors='none',zorder=10)
plt.title('City Locations colored by Population Rank')
plt.show()
